# wrench.waelstow.py
from unittest import TestCase

def list_tests(suites):
    """Takes a list of suites and returns an iterable of all the tests in
    the suites and their children.

    :param suites:
        A single or an interable of :class:`TestSuite` objects whose contents
        will be listed
    :returns:
        Iterable of :class:`TestCase` objects contained within the suites
    """
    for test in suites:
        if test.__class__.__name__ == 'ModuleImportFailure':
            # ignore the import fail module
            pass
        elif isinstance(test, TestCase):
            yield test
        else:
            for t in list_tests(test):
                yield t


def find_shortcut_tests(suites, shortcut_labels):
    """Takes a suite of tests and returns a list of tests that conform to the
    passed in list of short-cut labels.  A short-cut label begins with an "="
    and is then checked against the full test case and method name, if the
    shortcut string is in the full test name then the test is returned.

    Example:

    .. code-block::python

    >>> list_tests(suite)
    [
        test_foo (wrench.test.SomeTest),
        test_foo (wrench.test.AnotherTest),
        test_bar (wrench.test.AnotherTest),
        test_bar (wrench.test.DifferentTest),
    ]
    >>> find_shorcut_tests(suite, ['=foo'])
    [
        test_foo (wrench.test.SomeTest),
        test_foo (wrench.test.AnotherTest),
    ]
    >>> find_shorcut_tests(suite, ['=Another'])
    [
        test_foo (wrench.test.AnotherTest),
        test_bar (wrench.test.AnotherTest),
    ]

    :param suites:
        A single or iterable of :class:`TestSuite` objects to create the test
        subset from
    :param shortcut_labels:
        A list of short-cut labels to use to cull the list
    :returns:
        A list of :class:`TestCase` objects 
    """
    # strip the '=' from the front of each label
    labels = [label[1:] for label in shortcut_labels]

    results = []
    for test in list_tests(suites):
        name = '%s.%s.%s' % (test.__module__, test.__class__.__name__,
            test._testMethodName)
        for label in labels:
            if label in name:
                results.append(test)

    return results



# SizeRotatingFileHandler
#
# Based on:
#
# https://pypi.python.org/pypi/ConcurrentLogHandler
# version = 0.9.1
#
# Original work:
#
#     Copyright 2013 Lowell Alleman
#
#      Licensed under the Apache License, Version 2.0 (the "License"); you may
#      not use this file except in compliance with the License. You may obtain
#      a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#      WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#      License for the specific language governing permissions and limitations
#      under the License.

import os, sys, time
from random import randint
from logging import Handler, LogRecord
from logging.handlers import BaseRotatingHandler

try:
    import codecs
except ImportError:
    codecs = None

from portalocker import lock, unlock, LOCK_EX


# Workaround for handleError() in Python 2.7+ where record is written to stderr
class NullLogRecord(LogRecord):
    def __init__(self):
        pass
    def __getattr__(self, attr):
        return None

class SizeRotatingFileHandler(BaseRotatingHandler):
    """Rotating file handler that handles concurrent access without losing
    logs.  When a size limit is hit the log is rotated, with the old log
    renamed based on timestamp.

    Depending on the platform, the size limit may not be completely respected
    if several processes have it open simultaneously as it may need to wait
    for a process to let go.

    This handler is based on the version 0.9.1 of the 
    `ConcurrentLogHandler <https://pypi.python.org/pypi/ConcurrentLogHandler>`_
    by Lowell Alleman and with the exception of the removal of the "backup
    count" feature and how the files are renamed is a direct copy, the same
    caveats apply.
    """

    def __init__(self, filename, mode='a', maxBytes=0, encoding=None, 
            debug=True, delay=0):
        """
        Open the specified file and use it as the stream for logging.

        By default, the file grows indefinitely. You can specify particular
        values of maxBytes to allow the file to rollover at a predetermined
        size.

        Rollover occurs whenever the current log file is nearly maxBytes in
        length.  Old files are renamed based on timestamps.

        If maxBytes is zero, rollover never occurs.
        """
        # Absolute file name handling done by FileHandler since Python 2.5  
        BaseRotatingHandler.__init__(self, filename, mode, encoding, delay)
        self.delay = delay
        self._rotateFailed = False
        self.maxBytes = maxBytes
        self._open_lockfile()
        # For debug mode, swap out the "_degrade()" method with a more a verbose one.
        if debug:
            self._degrade = self._degrade_debug
    
    def _open_lockfile(self):
        # Use 'file.lock' and not 'file.log.lock' (Only handles the normal "*.log" case.)
        if self.baseFilename.endswith(".log"):
            lock_file = self.baseFilename[:-4]
        else:
            lock_file = self.baseFilename
        lock_file += ".lock"
        self.stream_lock = open(lock_file,"w")
    
    def _open(self, mode=None):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        
        Note:  Copied from stdlib.  Added option to override 'mode'
        """
        if mode is None:
            mode = self.mode
        if self.encoding is None:
            stream = open(self.baseFilename, mode)
        else:
            stream = codecs.open(self.baseFilename, mode, self.encoding)
        return stream
    
    def _close(self):
        """ Close file stream.  Unlike close(), we don't tear anything down, we
        expect the log to be re-opened after rotation."""
        if self.stream:
            try:
                if not self.stream.closed:
                    # Flushing probably isn't technically necessary, but it feels right
                    self.stream.flush()
                    self.stream.close()
            finally:
                self.stream = None
    
    def acquire(self):
        """ Acquire thread and file locks.  Re-opening log for 'degraded' mode.
        """
        # handle thread lock
        Handler.acquire(self)
        # Issue a file lock.  (This is inefficient for multiple active threads
        # within a single process. But if you're worried about high-performance,
        # you probably aren't using this log handler.)
        if self.stream_lock:
            # If stream_lock=None, then assume close() was called or something
            # else weird and ignore all file-level locks.
            if self.stream_lock.closed:
                # Daemonization can close all open file descriptors, see
                # https://bugzilla.redhat.com/show_bug.cgi?id=952929
                # Try opening the lock file again.  Should we warn() here?!?
                try:
                    self._open_lockfile()
                except Exception:
                    self.handleError(NullLogRecord())
                    # Don't try to open the stream lock again
                    self.stream_lock = None
                    return
            lock(self.stream_lock, LOCK_EX)
        # Stream will be opened as part by FileHandler.emit()

    def release(self):
        """ Release file and thread locks. If in 'degraded' mode, close the
        stream to reduce contention until the log files can be rotated. """
        try:
            if self._rotateFailed:
                self._close()
        except Exception:
            self.handleError(NullLogRecord())
        finally:
            try:
                if self.stream_lock and not self.stream_lock.closed:
                    unlock(self.stream_lock)
            except Exception:
                self.handleError(NullLogRecord())
            finally:
                # release thread lock
                Handler.release(self)
    
    def close(self):
        """
        Close log stream and stream_lock. """
        try:
            self._close()
            if not self.stream_lock.closed:
                self.stream_lock.close()
        finally:
            self.stream_lock = None
            Handler.close(self)
    
    def _degrade(self, degrade, msg, *args):
        """ Set degrade mode or not.  Ignore msg. """
        self._rotateFailed = degrade
        del msg, args   # avoid pychecker warnings
    
    def _degrade_debug(self, degrade, msg, *args):
        """ A more colorful version of _degade(). (This is enabled by passing
        "debug=True" at initialization).
        """
        if degrade:
            if not self._rotateFailed:
                sys.stderr.write("Degrade mode - ENTERING - (pid=%d)  %s\n" %
                                 (os.getpid(), msg % args))
                self._rotateFailed = True
        else:
            if self._rotateFailed:
                sys.stderr.write("Degrade mode - EXITING  - (pid=%d)   %s\n" %
                                 (os.getpid(), msg % args))
                self._rotateFailed = False
    
    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        self._close()

        try:
            dfn = self.baseFilename + "." + time.strftime(self.SUFFIX, 
                time.gmtime(int(time.time())))

            # Determine if we can rename the log file or not. Windows refuses
            # to rename an open file, Unix is inode base so it doesn't care.
            
            # Attempt to rename logfile to tempname:  There is a slight
            # race-condition here, but it seems unavoidable
            tmpname = None
            while not tmpname or os.path.exists(tmpname):
                tmpname = "%s.rotate.%08d" % (self.baseFilename, randint(0,99999999))
            try:
                # Do a rename test to determine if we can successfully rename
                # the log file
                os.rename(self.baseFilename, tmpname)
            except (IOError, OSError):
                exc_value = sys.exc_info()[1]
                self._degrade(True, "rename failed.  File in use?  "
                              "exception=%s", exc_value)
                return

            os.rename(tmpname, dfn)
            #print "%s -> %s" % (self.baseFilename, dfn)
            self._degrade(False, "Rotation completed")
        finally:
            # Re-open the output stream, but if "delay" is enabled then wait
            # until the next emit() call. This could reduce rename contention in
            # some usage patterns.
            if not self.delay:
                self.stream = self._open()
    
    def shouldRollover(self, record):
        """
        Determine if rollover should occur.

        For those that are keeping track. This differs from the standard
        library's RotatingLogHandler class. Because there is no promise to keep
        the file size under maxBytes we ignore the length of the current record.
        """
        del record  # avoid pychecker warnings
        # Is stream is not yet open, skip rollover check. (Check will occur on
        # next message, after emit() calls _open())
        if self.stream is None:
            return False
        if self._shouldRollover():
            # If some other process already did the rollover (which is possible
            # on Unix) the file our stream may now be named "log.1", thus
            # triggering another rollover. Avoid this by closing and opening
            # "log" again.
            self._close()
            self.stream = self._open()
            return self._shouldRollover()
        return False
    
    def _shouldRollover(self):
        if self.maxBytes > 0:                   # are we rolling over?
            self.stream.seek(0, 2)  #due to non-posix-compliant Windows feature
            if self.stream.tell() >= self.maxBytes:
                return True
            else:
                self._degrade(False, "Rotation done or not needed at this time")
        return False


# Publish this class to the "logging.handlers" module so that it can be use 
# from a logging config file via logging.config.fileConfig().
import logging.handlers
logging.handlers.SizeRotatingFileHandler = SizeRotatingFileHandler



import os, logging
from functools import wraps

# =============================================================================
# Constants
# =============================================================================

#: Log formatting for vt100 with function name highlighted and message
LOG_FORMAT_ESCAPED = '\033[1m%(funcName)s\033[0m: %(message)s'

#: Log format for files with date/time, function info, pid/tid and message
LOG_FORMAT_STANDARD = ('%(asctime)s %(name)s.%(funcName)s: '
    'pid:tid=%(process)d:%(thread)d %(message)s')

# =============================================================================
# Logging Configuration
# =============================================================================

def configure_file_logger(name, log_dir, log_level=logging.DEBUG):
    """Configures logging to use the :class:`SizeRotatingFileHandler`"""
    from .srothandler import SizeRotatingFileHandler

    root = logging.getLogger()
    root.setLevel(log_level)
    handler = SizeRotatingFileHandler(os.path.join(log_dir, '%s.log' % name))
    handler.setLevel(log_level)
    handler.setFormatter(logging.Formatter(LOG_FORMAT_STANDARD))

    root.addHandler(handler)


def configure_stdout_logger(log_level=logging.DEBUG):
    """Configures logging to use STDOUT"""
    root = logging.getLogger()
    root.setLevel(log_level)

    handler = logging.StreamHandler()
    handler.setLevel(log_level)
    handler.setFormatter(logging.Formatter(LOG_FORMAT_ESCAPED))

    root.addHandler(handler)


def default_logging_dict(log_dir, handlers=['file'], filename='debug.log'):
    """Returns a logging configuration dictionary with reasonable defaults.
    Defines two handlers: "default" and "file", that go to STDOUT and a
    :class:`SizeRotatingFileHandler`.

    :param log_dir:
        Directory for logs to go into.
    :param handlers:
        Which logging handlers to use.  Defaults to only 'file'
    :param filename:
        Base name of the file to log to.  Defaults to "debug.log".
    """
    d = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'escaped': {
                'format':LOG_FORMAT_ESCAPED,
            },
            'standard': {
                'format':LOG_FORMAT_STANDARD,
            },
        },
        'handlers': {
            'default': {
                'level':'DEBUG',
                'class':'logging.StreamHandler',
                'formatter':'escaped',
            },
            'file': {
                'level':'DEBUG',
                'class':'wrench.logtools.srothandler.SizeRotatingFileHandler',
                'filename': os.path.abspath(os.path.join(log_dir, filename)),
                'formatter':'standard',
                'maxBytes':300000,
            },
        },
        'loggers': {
            '': {
                'handlers':handlers,
                'propagate': False,
                'level':'DEBUG',
            },
        },
    }

    return d

# ============================================================================
# Decorators
# ============================================================================

def silence_logging(method):
    """Disables logging for the duration of what is being wrapped.  This is
    particularly useful when testing if a test method is supposed to issue an
    error message which is confusing that the error shows for a successful
    test.

    .. warning::

        Normally you can use function decorators to wrap a class, but if you
        wrap a :class:`TestCase` class what is returned is a function which
        means it won't be found by the test discovery process.  Wrap the
        individual test case methods, or alter logging in the
        :class:`TestCase.setUp` and :class`TestCase.tearDown` methods.
    """
    @wraps(method)
    def wrapper(*args, **kwargs):
        logging.disable(logging.ERROR)
        result = method(*args, **kwargs)
        logging.disable(logging.NOTSET)
        return result
    return wrapper



import os, shutil, tempfile
from datetime import datetime, date, time
from unittest import TestCase, TestSuite, defaultTestLoader

from wrench.utils import (ExtendedEnum, When, dynamic_load, 
    camelcase_to_underscore, rows_to_columns, temp_directory, 
    replaced_directory)
from wrench.waelstow import list_tests, find_shortcut_tests

# =============================================================================

class ColourEnum(ExtendedEnum):
    red = 'r'
    blue = 'b'
    green = 'g'


class TestEnum(TestCase):
    def test_enum(self):
        # test choices()
        choices = ColourEnum.choices()
        self.assertEqual(len(choices), 3)
        self.assertIn(('r', 'red'), choices)
        self.assertIn(('b', 'blue'), choices)
        self.assertIn(('g', 'green'), choices)

        # test count()
        self.assertEqual(3, ColourEnum.count())

        # test values()
        values = ColourEnum.values()
        self.assertEqual(len(values), 3)
        self.assertIn('r', values)
        self.assertIn('b', values)
        self.assertIn('g', values)

        # test names()
        names = ColourEnum.names()
        self.assertEqual(len(names), 3)
        self.assertIn('red', names)
        self.assertIn('blue', names)
        self.assertIn('green', names)

        
class TestWhen(TestCase):
    def setUp(self):
        self.full_date = datetime(1972, 1, 31, 13, 55)
        self.zero_date = datetime(1972, 1, 31, 0, 0)
        self.only_time = time(13, 55)
        self.epoch = 65732100
        self.mepoch = self.epoch * 1000
        self.micro_epoch = '%s.0000' % (self.epoch)

    def test_constructor(self):
        self.assertEqual(self.full_date, When(datetime=self.full_date).datetime)
        self.assertEqual(self.full_date, 
            When(date=date(1972, 1, 31), time=time(13, 55)).datetime)
        self.assertEqual(self.full_date, 
            When(date=date(1972, 1, 31), time_string='13:55').datetime)
        self.assertEqual(self.full_date, 
            When(date=date(1972, 1, 31), time_string='13:55:00').datetime)

        self.assertEqual(self.full_date, When(epoch=self.epoch).datetime)
        self.assertEqual(self.full_date, When(milli_epoch=self.mepoch).datetime)

        # default time
        self.assertEqual(self.zero_date, When(date=date(1972, 1, 31)).datetime)

        # time only
        self.assertEqual(self.only_time, When(time=time(13, 55)).time)
        self.assertEqual(self.only_time, When(time_string='13:55').time)

        # parse detection
        self.assertEqual(self.zero_date, When(detect='1972-01-31').datetime)
        self.assertEqual(self.only_time, When(detect='13:55').time)
        self.assertEqual(self.only_time, When(detect='13:55:00').time)
        self.assertEqual(self.full_date, 
            When(detect='1972-01-31 13:55').datetime)
        self.assertEqual(self.full_date, 
            When(detect='1972-01-31 13:55:00').datetime)
        self.assertEqual(self.full_date, 
            When(detect='1972-01-31T13:55Z').datetime)
        self.assertEqual(self.full_date, 
            When(detect='1972-01-31T13:55:00Z').datetime)
        self.assertEqual(self.full_date, 
            When(detect='1972-01-31T13:55:00.00Z').datetime)

        # parse force
        self.assertEqual(self.zero_date, When(parse_date='1972-01-31').datetime)
        self.assertEqual(self.only_time, When(parse_time='13:55').time)
        self.assertEqual(self.only_time, When(parse_time_sec='13:55:00').time)
        self.assertEqual(self.full_date, 
            When(parse_datetime='1972-01-31 13:55').datetime)
        self.assertEqual(self.full_date, 
            When(parse_datetime_sec='1972-01-31 13:55:00').datetime)
        self.assertEqual(self.full_date, 
            When(parse_datetime_utc='1972-01-31T13:55Z').datetime)
        self.assertEqual(self.full_date, 
            When(parse_datetime_sec_utc='1972-01-31T13:55:00Z').datetime)
        self.assertEqual(self.full_date, 
            When(parse_iso_micro='1972-01-31T13:55:00.00Z').datetime)

        # check errors
        with self.assertRaises(ValueError):
            When(detect='abc')

        with self.assertRaises(ValueError):
            When(parse_date='abc')

        with self.assertRaises(ValueError):
            When(parse_time='abc')

        with self.assertRaises(ValueError):
            When(parse_time_sec='abc')

        with self.assertRaises(ValueError):
            When(parse_datetime='abc')

        with self.assertRaises(ValueError):
            When(parse_datetime_sec='abc')

        with self.assertRaises(ValueError):
            When(parse_datetime_utc='abc')

        with self.assertRaises(ValueError):
            When(parse_datetime_sec_utc='abc')

        with self.assertRaises(ValueError):
            When(parse_iso_micro='abc')

        with self.assertRaises(AttributeError):
            When()

    def test_values(self):
        when = When(datetime=self.full_date)

        self.assertEqual('1972-01-31', when.string.date)
        self.assertEqual('13:55', when.string.time)
        self.assertEqual('13:55:00', when.string.time_sec)
        self.assertEqual('1972-01-31 13:55', when.string.datetime)
        self.assertEqual('1972-01-31 13:55:00', when.string.datetime_sec)
        self.assertEqual('1972-01-31T13:55Z', when.string.datetime_utc)
        self.assertEqual('1972-01-31T13:55:00Z', when.string.datetime_sec_utc)
        self.assertEqual('1972-01-31T13:55:00.0Z', when.string.iso_micro)

        self.assertEqual(self.full_date, when.datetime)
        self.assertEqual(self.full_date.date(), when.date)
        self.assertEqual(self.only_time, when.time)
        self.assertEqual(self.epoch, when.epoch)
        self.assertEqual(self.mepoch, when.milli_epoch)


class TestUtils(TestCase):
    def test_dynamic_load(self):
        import os
        fn = dynamic_load('os.path.abspath')
        self.assertEqual(os.path.abspath, fn)

    def test_camelcase(self):
        pairs = [
            ('one', 'one'),
            ('one_two', 'oneTwo'),
            ('one_two', 'OneTwo'),
        ]

        for pair in pairs:
            self.assertEqual(pair[0], camelcase_to_underscore(pair[1]))

    def test_rows_to_cols(self):
        matrix = [ 
            [0, 1, 2],
            [3, 4, 5],
            [6, 7, 8],
        ]

        expected = [ 
            [0, 3, 6],
            [1, 4, 7],
            [2, 5, 8],
        ]

        self.assertEqual(expected, rows_to_columns(matrix))


class TestContexts(TestCase):
    def test_temp_dir(self):
        created_td = ''
        with temp_directory() as td:
            created_td = td
            name = os.path.join(td, 'a.txt')
            with open(name, 'w') as f:
                f.write('foo')

            os.path.isfile(name)

        # verify that the temp directory has been cleaned up
        self.assertFalse(os.path.exists(created_td))

        # -- test failure still cleans up
        try:
            with temp_directory() as td:
                created_td = td
                raise RuntimeError()
        except:
            pass

        self.assertFalse(os.path.exists(created_td))

    def test_replace_dir(self):
        # create a temp directory and put something in it which is to be
        # replaced
        test_dir = tempfile.mkdtemp()
        orig_file = os.path.join(test_dir, 'a.txt')
        replace_file = os.path.join(test_dir, 'b.txt')

        with open(orig_file, 'w') as f:
            f.write('foo')

        # test not a dir handling
        with self.assertRaises(AttributeError):
            # call context manager by hand as putting it in a "with" will
            # result in unreachable code which blows our testing coverage
            rd = replaced_directory(orig_file)
            rd.__enter__()

        # replace_directory should handle trailing slashes
        test_dir += '/'

        created_td = ''
        with replaced_directory(test_dir) as td:
            created_td = td

            # put something in the replaced directory
            with open(replace_file, 'w') as f:
                f.write('bar')

            # original should be moved out of path, replaced should exist
            self.assertFalse(os.path.exists(orig_file))
            self.assertTrue(os.path.exists(replace_file))

        # original should be back, replaced should be gone
        self.assertTrue(os.path.exists(orig_file))
        self.assertFalse(os.path.exists(replace_file))
        self.assertFalse(os.path.exists(created_td))

        # -- test failure still cleans up
        try:
            with replaced_directory(test_dir) as td:
                created_td = td
                raise RuntimeError()
        except:
            pass

        self.assertTrue(os.path.exists(orig_file))
        self.assertFalse(os.path.exists(created_td))

        # -- cleanup testcase
        shutil.rmtree(test_dir)


class WaelstowTest(TestCase):
    def setUp(self):
        class ATestCase(TestCase):
            def test_common(self):
                pass

            def test_a(self):
                pass

        class BTestCase(TestCase):
            def test_common(self):
                pass

            def test_b(self):
                pass

        class CTestCase(TestCase):
            def test_common(self):
                pass

        # mock up what python.unittests does for failures make sure it is
        # ignored properly
        class ModuleImportFailure(TestCase):
            pass

        self.suite_a = defaultTestLoader.loadTestsFromTestCase(ATestCase)
        self.suite_b = defaultTestLoader.loadTestsFromTestCase(BTestCase)
        self.suite_c = defaultTestLoader.loadTestsFromTestCase(CTestCase)

        group = TestSuite([self.suite_a, self.suite_b])
        self.suite = TestSuite([self.suite_c, group, ModuleImportFailure()])

    def assert_test_strings(self, expected, tests):
        names = [str(test) for test in tests]
        self.assertEqual(set(expected), set(names))

    def test_list_tests(self):
        expected = [
            'test_common (tests.ATestCase)',
            'test_a (tests.ATestCase)',
            'test_common (tests.BTestCase)',
            'test_b (tests.BTestCase)',
            'test_common (tests.CTestCase)',
        ]

        tests = list(list_tests(self.suite))
        self.assert_test_strings(expected, tests)

    def test_find_shortcuts(self):
        expected = [
            'test_common (tests.ATestCase)',
            'test_common (tests.BTestCase)',
            'test_common (tests.CTestCase)',
        ]

        tests = find_shortcut_tests(self.suite, ['=common'])
        self.assert_test_strings(expected, tests)

        expected = [
            'test_a (tests.ATestCase)',
            'test_b (tests.BTestCase)',
        ]

        tests = find_shortcut_tests(self.suite, ['=_a', '=_b'])
        self.assert_test_strings(expected, tests)

        expected = [
            'test_a (tests.ATestCase)',
            'test_common (tests.ATestCase)',
        ]

        tests = find_shortcut_tests(self.suite, ['=ATest'])
        self.assert_test_strings(expected, tests)

    def test_misc(self):
        # misc stuff to hit our 100% coverage 
        for t in list_tests(self.suite):
            t.run()



import os, sys, copy
import logging.config
from six import StringIO
from unittest import TestCase

from wrench.utils import temp_directory
from wrench.logtools.utils import (configure_file_logger, 
    configure_stdout_logger, default_logging_dict, silence_logging)

# =============================================================================

class GotHere(Exception):
    pass


@silence_logging
class ShouldBeQuiet(object):
    def logthing(self):
        logger = logging.getLogger(__name__)
        logger.debug('this is a message')
        raise GotHere()


@silence_logging
def should_be_quiet():
    logger = logging.getLogger(__name__)
    logger.debug('this is a message')
    raise GotHere()

# =============================================================================

class TestLogging(TestCase):
    def test_config_file_logger(self):
        config = copy.deepcopy(logging.Logger.manager.loggerDict)

        with temp_directory() as td:
            configure_file_logger('debug', td)
            logger = logging.getLogger(__name__)
            logger.debug('this is a message')

            logfile = os.path.abspath(os.path.join(td, 'debug.log'))
            self.assertTrue(os.path.exists(logfile))

        logging.Logger.manager.loggerDict = config

    def test_config_stdout_logger(self):
        config = copy.deepcopy(logging.Logger.manager.loggerDict)
        saved_stderr = sys.stderr
        try:
            out = StringIO()
            sys.stderr = out
            configure_stdout_logger()
            logger = logging.getLogger(__name__)
            logger.debug('this is a message')

            output = out.getvalue().strip()
            self.assertIn('this is a message', output)

        finally:
            sys.stderr = saved_stderr
        logging.Logger.manager.loggerDict = config

    def test_default_logging_dict(self):
        config = copy.deepcopy(logging.Logger.manager.loggerDict)

        with temp_directory() as td:
            d = default_logging_dict(td)
            logging.config.dictConfig(d)

            logger = logging.getLogger(__name__)
            logger.debug('this is a message')

            logfile = os.path.abspath(os.path.join(td, 'debug.log'))
            self.assertTrue(os.path.exists(logfile))

        logging.Logger.manager.loggerDict = config

    def test_silence_decorator(self):
        config = copy.deepcopy(logging.Logger.manager.loggerDict)
        saved_stderr = sys.stderr
        try:
            out = StringIO()
            sys.stderr = out

            with self.assertRaises(GotHere):
                should_be_quiet()

            s = ShouldBeQuiet()
            with self.assertRaises(GotHere):
                s.logthing()

            output = out.getvalue().strip()
            self.assertEqual(0, len(output))

        finally:
            sys.stderr = saved_stderr
        logging.Logger.manager.loggerDict = config



import contextlib, os, shutil, sys, tempfile
from datetime import datetime, time
import time as time_mod

from enum import Enum

# =============================================================================

# compensate for long disappearing in py3
if sys.version_info > (3,):
    long = int

# =============================================================================
# Extended Enum 
# =============================================================================

class ExtendedEnum(Enum):
    """Extends ``Enum`` (either native or enum34.enum depending on python
    version) and adds some features.
    """

    @classmethod
    def choices(cls):
        """Returns a "choices" list of tuples.  Each member of the list is one
        of the possible items in the ``Enum``, with the first item in the pair
        being the value and the second the name.  This is compatible with
        django's choices attribute in fields.

        :returns:
            List of tuples
        """
        result = []
        for name, member in cls.__members__.items():
            result.append((member.value, name))

        return result

    @classmethod
    def count(cls):
        """Returns the number of items in the enumeration."""
        return len(cls.__members__)

    @classmethod
    def values(cls):
        """Returns a list of the values of the items in the enumeration."""
        return [e.value for e in list(cls)]

    @classmethod
    def names(cls):
        """Returns a list of the names of the items in the enumeration."""
        return [e.name for e in list(cls)]

# =============================================================================
# Date Conversion
# =============================================================================

class TimeOnlyError(Exception):
    """Exception indicating that a date operation was attempted on a
    :class:`When` object that only wraps a python ``time`` instance."""
    pass


class When(object):
    # https://docs.python.org/2/library/datetime.html#strftime-strptime-behavior
    """Date/time conversion utility.  A When object wraps either a python
    ``datetime`` or ``time`` instance, providing a common mechanism for
    building and converting them into different formats.

    If When is wrapping a ``time`` object then some methods will raise a
    :class:`TimeOnlyError`.

    .. _when-formats:

    Supported formats for parsing and display::

        'date': '%Y-%m-%d'
        'time': '%H:%M'
        'time_sec': '%H:%M:%S'
        'datetime': '%Y-%m-%d %H:%M'
        'datetime_sec': '%Y-%m-%d %H:%M:%S'
        'datetime_utc': '%Y-%m-%dT%H:%MZ'
        'datetime_sec_utc': '%Y-%m-%dT%H:%M:%SZ'
        'iso_micro': '%Y-%m-%dT%H:%M:%S.%fZ'

    .. warning::

        All python ``datetime`` objects in this class are naive and have no
        timezone information associated with them, this includes various
        formats labelled "utc".
    """
    parse_formats = {
        'date': '%Y-%m-%d',
        'time': '%H:%M',
        'time_sec': '%H:%M:%S',
        'datetime': '%Y-%m-%d %H:%M',
        'datetime_sec': '%Y-%m-%d %H:%M:%S',
        'datetime_utc': '%Y-%m-%dT%H:%MZ',
        'datetime_sec_utc': '%Y-%m-%dT%H:%M:%SZ',
        'iso_micro': '%Y-%m-%dT%H:%M:%S.%fZ',
    }

    class WhenStrformat(object):
        def __init__(self, when):
            self.when = when

        def __getattribute__(self, name):
            if name == 'when':
                return object.__getattribute__(self, name)

            formatter = self.when.parse_formats[name]
            if name == 'iso_micro':
                # iso_micro needs special handling as python's %f returns a
                # zero padded number but the standard doesn't pad the value
                formatter = self.when.parse_formats['datetime_sec_utc']

            result = self.when.datetime.strftime(formatter)

            if name == 'iso_micro':
                result = result[:-1] + '.%dZ' % int(
                    self.when.datetime.strftime('%f'))

            return result


    def __init__(self, **kwargs):
        """Create a When object.  The constructor accepts a variety of
        keywords depending on what type of date or time information you wish
        to convert.  Most keywords result in a python ``datetime`` object
        being wrapped, but certain cases use a ``time`` object.  If the When
        is only a ``time`` wrapper then some of the methods will not be
        available.

        :param datetime:
            Create ``When`` using a python ``datetime`` object
        :param date:
            Create ``When`` using a python ``date`` object, can be used in
            conjunction with the ``time`` keyword.  If used without the
            ``time`` keyword the time portion will be set to midnight.
        :param time:
            Create ``When`` using a python ``time`` object.  If used on its
            own the date based methods of ``When`` will not be allowed.  This
            keyword can be used in conjunction with the ``date`` keyword to
            create a fully qualified ``When``.
        :param time_string:
            Create ``When`` using a string containing time information.
            Handles either 'hour:minute' or 'hour:minute:second'.  Like the
            ``time`` keyword, can be used in conjunction with ``date``.
        :param epoch:
            Create ``When`` using an integer epoch value
        :param milli_epoch:
            Create ``When`` using an integer that is 1000 * an epoch value with
            the last thousands being milli-epoch.
        :param detect:
            Create ``When`` by parsing a string which is compared against the
            list of available string parsers.
        :param parse_*: 
            Create ``When`` by parsing a string using the specific 
            :ref:`Supported formats <when-formats>` given.
            For example, ``parse_iso_micro`` expects the ISO 8601 format.

        :raises ValueError:
            If a bad string is passed to ``detect`` or ``parse_*`` keywords
        :raises AttributeError:
            If the constructor was called without sufficient arguments to
            result in a date or time being wrapped.
        """
        self._datetime = None
        self._time = None

        if 'datetime' in kwargs:
            self._datetime = kwargs['datetime']
        elif 'date' in kwargs:
            d = kwargs['date']
            if 'time' in kwargs:
                t = kwargs['time']
            elif 'time_string' in kwargs:
                t = self._parse_time_string(kwargs['time_string'])
            else:
                t = time()

            self._datetime = datetime(d.year, d.month, d.day, t.hour, t.minute,
                t.second)
        elif 'time' in kwargs:
            self._time = kwargs['time']
        elif 'time_string' in kwargs:
            self._time = self._parse_time_string(kwargs['time_string'])
        elif 'epoch' in kwargs:
            self._datetime = datetime.fromtimestamp(kwargs['epoch'])
        elif 'milli_epoch' in kwargs:
            mepoch = kwargs['milli_epoch']
            epoch = float(mepoch) / 1000.0
            milli = float(mepoch) - (epoch * 1000.0)
            self._datetime = datetime.fromtimestamp(long(epoch))
            self._datetime.replace(microsecond = int(milli / 10.0))
        elif 'detect' in kwargs:
            for f in self.parse_formats.values():
                try:
                    self._datetime = datetime.strptime(kwargs['detect'], f)
                    break
                except ValueError:
                    # couldn't parse using this format, ignore and try again
                    pass

            if not self._datetime:
                # nothing parsed
                raise ValueError(('could not parse the date/time passed to '
                    'detect'))
        else:
            # loop through all the possible kwargs looking for parse_* keys,
            # if found parse based on that and stop
            for key, value in kwargs.items():
                if key.startswith('parse_'):
                    f = self.parse_formats[key[6:]]
                    if key.startswith('parse_time'):
                        self._time = self._parse_time_string(value)
                    else:
                        self._datetime = datetime.strptime(value, f)

                    break

        if not self._datetime and not self._time:
            raise AttributeError('invalid keyword arguments')

    def _parse_time_string(self, value):
        parts = value.split(':')
        parts = [int(p) for p in parts]
        return time(*parts)

    @property
    def string(self):
        """Returns a placeholder object that has an attribute for each one of
        the :ref:`Supported formats <when-formats>`.  

        Example::

            >>> When(datetime=d).string.iso_micro
            1972-01-31T13:45:00.2

        """
        return When.WhenStrformat(self)

    @property
    def date(self):
        """Returns a python ``date`` object."""
        return self._datetime.date()

    @property
    def datetime(self):
        """Returns a python ``datetime`` object."""
        return self._datetime

    @property
    def time(self):
        """Returns a python ``time`` object."""
        if self._datetime:
            return self._datetime.time()

        return self._time

    @property
    def epoch(self):
        """Returns an integer version of epoch, i.e. the number of seconds
        since Jan 1, 1970."""
        return long(time_mod.mktime(self._datetime.timetuple()))

    @property
    def milli_epoch(self):
        """Returns an int of the epoch * 1000 + milliseconds."""
        return self.epoch * 1000 + self._datetime.microsecond * 10

# =============================================================================
# Utility Methods
# =============================================================================

def dynamic_load(name):
    """Equivalent of "from X import Y" statement using dot notation to specify
    what to import and return.  For example, foo.bar.thing returns the item
    "thing" in the module "foo.bar" """
    pieces = name.split('.')
    item = pieces[-1]
    mod_name = '.'.join(pieces[:-1])

    mod = __import__(mod_name, globals(), locals(), [item])
    return getattr(mod, item)


def camelcase_to_underscore(text):
    prev_cap = text[0].isupper()
    result = [text[0].lower(), ]
    for letter in text[1:]:
        if letter.isupper():
            if not prev_cap:
                result.append('_')

            result.append(letter.lower())
            prev_cap = True
        else:
            result.append(letter)
            prev_cap = False

    return ''.join(result)


def rows_to_columns(matrix):
    """Takes a two dimensional array and returns an new one where rows in the
    first become columns in the second."""
    num_rows = len(matrix)
    num_cols = len(matrix[0])

    data = []
    for i in range(0, num_cols):
        data.append([matrix[j][i] for j in range(0, num_rows)])

    return data

# =============================================================================
# Contexts
# =============================================================================

@contextlib.contextmanager
def replaced_directory(dirname):
    """This ``Context`` is used to move the contents of a directory elsewhere
    temporarily and put them back upon exit.  This allows testing code to use
    the same file directories as normal code without fear of damage.

    The name of the temporary directory which contains your files is yielded.

    :param dirname:
        Path name of the directory to be replaced.
    """
    if dirname[-1] == '/':
        dirname = dirname[:-1]

    full_path = os.path.abspath(dirname)
    if not os.path.isdir(full_path):
        raise AttributeError('dir_name must be a directory')

    base, name = os.path.split(full_path)

    # create a temporary directory, move provided dir into it and recreate the
    # directory for the user
    tempdir = tempfile.mkdtemp()
    shutil.move(full_path, tempdir)
    os.mkdir(full_path)
    try:
        yield tempdir

    finally:
        # done context, undo everything
        shutil.rmtree(full_path)
        moved = os.path.join(tempdir, name)
        shutil.move(moved, base)
        shutil.rmtree(tempdir)


@contextlib.contextmanager
def temp_directory():
    """This ``Context`` creates a temporary directory and yields its path.
    Upon exit the directory is removed.
    """
    tempdir = tempfile.mkdtemp()
    try:
        yield tempdir
    finally:
        shutil.rmtree(tempdir)
